@@ -1,6 +1,6 @@
 ;;; org-transclusion.el --- Transclude text content via links -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2021  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2022  Free Software Foundation, Inc.
 
 ;; This program is free software: you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by the
@@ -17,12 +17,12 @@
 
 ;; Author:        Noboru Ota <me@nobiot.com>
 ;; Created:       10 October 2020
-;; Last modified: 29 December 2021
+;; Last modified: 12 June 2022
 
 ;; URL: https://github.com/nobiot/org-transclusion
 ;; Keywords: org-mode, transclusion, writing
 
-;; Version: 1.1.0
+;; Version: 1.3.0
 ;; Package-Requires: ((emacs "27.1") (org "9.4"))
 
 ;; This file is not part of GNU Emacs.
@@ -43,9 +43,6 @@
 (require 'org-transclusion-font-lock)
 (require 'text-property-search)
 (require 'seq)
-(declare-function org-translusion-indent-add-properties
-                  "org-transclusion-indent-mode")
-(defvar org-indent-mode)
 
 ;;;; Customization
 
@@ -63,7 +60,7 @@ Intended for :set property for `customize'."
   (when (featurep 'org-transclusion)
     (org-transclusion-load-extensions-maybe 'force)))
 
- (defcustom org-transclusion-extensions '(org-transclusion-src-lines)
+ (defcustom org-transclusion-extensions '(org-transclusion-src-lines org-transclusion-font-lock)
   "Extensions to be loaded with org-transclusion.el."
   :set #'org-transclusion-set-extensions
   :type
@@ -104,6 +101,17 @@ See `display-buffer' for example options."
   :type '(choice (const :tag "No lighter" "") string)
   :safe 'stringp)
 
+(defcustom org-transclusion-after-add-functions nil
+  "Functions to be called after a transclusion content has been added.
+The hook runs after the content and the read-only text property
+have been added so it is not supposed to manipulate the content
+but to add further text properties.  For example, it is used by
+the `org-transclusion-indent-mode' extension to support
+`org-indent-mode'.  The functions are called with arguments beg
+and end, pointing to the beginning and end of the transcluded
+content."
+  :type '(repeat function))
+
 ;;;; Faces
 
 (defface org-transclusion-source-fringe
@@ -197,6 +205,7 @@ that consists of the following properties:
     org-transclusion-keyword-value-disable-auto
     org-transclusion-keyword-value-only-contents
     org-transclusion-keyword-value-exclude-elements
+    org-transclusion-keyword-value-expand-links
     org-transclusion-keyword-current-indentation)
   "Define a list of functions used to parse a #+transclude keyword.
 These functions take a single argument, the whole keyword value
@@ -218,6 +227,7 @@ regexp from the string.")
 
 (defvar org-transclusion-map
   (let ((map (make-sparse-keymap)))
+    (set-keymap-parent map org-mode-map)
     (define-key map (kbd "e") #'org-transclusion-live-sync-start)
     (define-key map (kbd "g") #'org-transclusion-refresh)
     (define-key map (kbd "d") #'org-transclusion-remove)
@@ -225,8 +235,6 @@ regexp from the string.")
     (define-key map (kbd "D") #'org-transclusion-demote-subtree)
     (define-key map (kbd "o") #'org-transclusion-open-source)
     (define-key map (kbd "O") #'org-transclusion-move-to-source)
-    (define-key map (kbd "TAB") #'org-cycle)
-    (define-key map (kbd "C-c C-c") #'org-ctrl-c-ctrl-c)
     map)
   "It is the local-map used within a transclusion.
 As the transcluded text content is read-only, these keybindings
@@ -271,9 +279,14 @@ specific keybindings; namely:
 (defmacro org-transclusion-with-inhibit-read-only (&rest body)
   "Run BODY with `'inhibit-read-only` t."
   (declare (debug t) (indent 0))
-  `(let* ((inhibit-read-only t))
-     (progn
-           ,@body)))
+  (let ((modified (make-symbol "modified")))
+    `(let* ((,modified (buffer-modified-p))
+            (inhibit-read-only t))
+       (unwind-protect
+           (progn
+             ,@body)
+         (unless ,modified
+           (restore-buffer-modified-p nil))))))
 
 ;;;; Commands
 
@@ -424,7 +437,7 @@ does not support all the elements.
                    "No content found with \"%s\".  Check the link at point %d, line %d"
                    (org-element-property :raw-link link) (point) (org-current-line))
                   nil))
-        (let ((beg (point))
+        (let ((beg (line-beginning-position))
               (end))
           (org-transclusion-with-inhibit-read-only
             (when (save-excursion
@@ -438,9 +451,7 @@ does not support all the elements.
               ;; `org-transclusion-keyword-remove' checks element at point is a
               ;; keyword or not
               (org-transclusion-keyword-remove)))
-          (when (and (featurep 'org-indent) org-indent-mode
-                     (memq 'org-transclusion-indent-mode org-transclusion-extensions))
-            (org-translusion-indent-add-properties beg end)))
+          (run-hook-with-args 'org-transclusion-after-add-functions beg end))
         t))))
 
 ;;;###autoload
@@ -517,11 +528,7 @@ When success, return the beginning point of the keyword re-inserted."
             ;; inevitably have the same position (location "between" lines)
             (when mkr-at-beg (move-marker mkr-at-beg beg))
             ;; Go back to the beginning of the inserted keyword line
-            (goto-char beg)
-            (when (and (featurep 'org-indent) org-indent-mode
-                       (memq 'org-transclusion-indent-mode
-                             org-transclusion-extensions))
-              (org-translusion-indent-add-properties beg (line-end-position))))
+            (goto-char beg))
           beg))
     (message "Nothing done. No transclusion exists here.") nil))
 
@@ -736,19 +743,18 @@ set in `before-save-hook'.  It also move the point back to
   "Remove transclusions before `kill-buffer' or `kill-emacs'.
 Intended to be used with `kill-buffer-hook' and `kill-emacs-hook'
 to clear the file of the transcluded text regions.  This function
-also flags the buffer modified and `save-buffer'.  Calling
-`save-buffer' after remove-all and live-sync ensures the clearing
-process to occur.  This is reqiured because during live-sync,
-some hooks that manage the clearing process are temporarily
-turned off (removed)."
+also flags the buffer modified and `save-buffer'.  Calling the
+second `org-transclusion-remove-all' ensures the clearing process
+to occur.  This is reqiured because during live-sync, some hooks
+that manage the clearing process are temporarily turned
+off (removed)."
   ;; Remove transclusions first. To deal with an edge case where transclusions
   ;; were added for a capture buffer -- e.g. `org-capture' or `org-roam-catpure'
   ;; --, check is done for `buffer-file-name' to see if there is a file visited
   ;; by the buffer. If a "temp" buffer, there is no file being visited.
   (when (and (org-transclusion-remove-all)
              (buffer-file-name))
-    (set-buffer-modified-p t)
-    (save-buffer)))
+    (org-transclusion-remove-all)))
 
 ;;;;---------------------------------------------------------------------------
 ;;;; Functions for Transclude Keyword
@@ -822,12 +828,23 @@ It needs to be set in
 `org-transclusion-keyword-value-functions'."
   (list :current-indentation (current-indentation)))
 
+(defun org-transclusion-keyword-value-expand-links (string)
+  "It is a utility function used converting a keyword STRING to plist.
+It is meant to be used by `org-transclusion-get-string-to-plist'.
+It needs to be set in
+`org-transclusion-keyword-value-functions'."
+  (when (string-match ":expand-links" string)
+    (list :expand-links
+          (org-strip-quotes (match-string 0 string)))))
+
 (defun org-transclusion-keyword-remove ()
   "Remove the keyword element at point.
 Returns t if successful.  It checks if the element at point is a
 keyword.  If not, returns nil."
   (let* ((elm (org-element-at-point))
-         (beg (org-element-property :begin elm))
+         ;; Ignore and keep affiliated keywords before #+transclusion
+         ;; Fix issue #115
+         (beg (org-element-property :post-affiliated elm))
          (end (org-element-property :end elm))
          (post-blank (org-element-property :post-blank elm)))
     (when (string= "keyword" (org-element-type elm))
@@ -842,6 +859,7 @@ keyword.  If not, returns nil."
         (disable-auto (plist-get plist :disable-auto))
         (only-contents (plist-get plist :only-contents))
         (exclude-elements (plist-get plist :exclude-elements))
+        (expand-links (plist-get plist :expand-links))
         (custom-properties-string nil))
     (setq custom-properties-string
           (dolist (fn org-transclusion-keyword-plist-to-string-functions
@@ -856,6 +874,7 @@ keyword.  If not, returns nil."
             (when only-contents (format " :only-contents"))
             (when exclude-elements (format " :exclude-elements \"%s\""
                                            exclude-elements))
+            (when expand-links (format " :expand-links"))
             custom-properties-string
             "\n")))
 
@@ -959,7 +978,8 @@ based on the following arguments:
             (setq content (buffer-string)))))
     (insert
      (run-hook-with-args-until-success
-      'org-transclusion-content-format-functions type content))
+      'org-transclusion-content-format-functions
+      type content (plist-get keyword-values :current-indentation)))
     (setq end (point))
     (setq end-mkr (set-marker (make-marker) end))
     (add-text-properties beg end
@@ -1008,7 +1028,7 @@ This function sssumes the buffer is an Org buffer."
         (push (org-element-property :level h) list)))
     (when list (seq-min list))))
 
-(defun org-transclusion-content-format-org (type content)
+(defun org-transclusion-content-format-org (type content _indent)
   "Format text CONTENT from source before transcluding.
 Return content modified (or unmodified, if not applicable).
 
@@ -1033,14 +1053,17 @@ content."
       ;; Return the temp-buffer's string
       (buffer-string)))))
 
-(defun org-transclusion-content-format (_type content)
+(defun org-transclusion-content-format (_type content indent)
   "Format text CONTENT from source before transcluding.
 Return content modified (or unmodified, if not applicable).
 
-This is the default one.  It only returns the content as is."
+This is the default one.  It only returns the content as is.
+
+INDENT is the number of current indentation of the #+transclude."
   (with-temp-buffer
     (insert content)
     ;; Return the temp-buffer's string
+    (set-left-margin (point-min)(point-max) indent)
     (buffer-string)))
 
 (defun org-transclusion-content-org-marker (marker plist)
@@ -1051,17 +1074,14 @@ work to
   (if (and marker (marker-buffer marker)
            (buffer-live-p (marker-buffer marker)))
       (progn
-        (let ((only-contents (plist-get plist :only-contents))
-              (exclude-elements
-               (org-transclusion-keyword-plist-to-exclude-elements plist)))
-          (with-current-buffer (marker-buffer marker)
-            (org-with-wide-buffer
-             (goto-char marker)
-             (if (org-before-first-heading-p)
-                 (org-transclusion-content-org-buffer-or-element
-                  nil only-contents exclude-elements)
+        (with-current-buffer (marker-buffer marker)
+          (org-with-wide-buffer
+           (goto-char marker)
+           (if (org-before-first-heading-p)
                (org-transclusion-content-org-buffer-or-element
-                'only-element only-contents exclude-elements))))))
+                nil plist)
+             (org-transclusion-content-org-buffer-or-element
+              'only-element plist)))))
     (message "Nothing done. Cannot find marker for the ID.")))
 
 (defun org-transclusion-content-org-link (link plist)
@@ -1074,23 +1094,18 @@ work to
     ;; search-option is present.
     (let* ((path (org-element-property :path link))
            (search-option (org-element-property :search-option link))
-           (buf (find-file-noselect path))
-           (only-contents (plist-get plist :only-contents))
-           (exclude-elements
-            (org-transclusion-keyword-plist-to-exclude-elements plist)))
+           (buf (find-file-noselect path)))
       (with-current-buffer buf
         (org-with-wide-buffer
          (if search-option
              (progn
                (org-link-search search-option)
                (org-transclusion-content-org-buffer-or-element
-                'only-element only-contents exclude-elements))
+                'only-element plist))
            (org-transclusion-content-org-buffer-or-element
-            nil only-contents exclude-elements)))))))
+            nil plist)))))))
 
-(defun org-transclusion-content-org-buffer-or-element (&optional only-element
-                                                                 only-contents
-                                                                 exclude-elements)
+(defun org-transclusion-content-org-buffer-or-element (only-element plist)
   "Return a list of playload for transclusion.
 Tis function assumes the point is at the beginning of the org
 element to transclude.
@@ -1105,19 +1120,9 @@ When ONLY-ELEMENT is non-nil, this function looks at only the element
 at point; if nil, the whole buffer.
 
 This function applies multiple filters on the Org elements before
-construting the payload based on relevant user options and
-optional arguments as below:
-
-EXCLUDE-ELEMENTS adds elements to be excluded onto user option
-`org-transclusion-exclude-elements'.  The user option applies
-globally, the optional arguments can be applied for each
-transcluion.
-
-First section
-
-ONLY-CONTENTS applies filter to remove headline titles of the
-subtree, extracting only sections (including paragraphs, tables,
-etc.)."
+construting the payload based on PLIST. It is the
+\"keyword-plist\" for the transclusion being worked on; each
+property controls the filter applied to the transclusion."
   (let* ((el (org-element-context))
          (type (when el (org-element-type el))))
     (if (or (not el)(not type))
@@ -1129,6 +1134,10 @@ etc.)."
         (setq el (org-element-property :parent el)))
       (let ((beg (org-element-property :begin el))
             (end (org-element-property :end el))
+            (only-contents (plist-get plist :only-contents))
+            (exclude-elements
+             (org-transclusion-keyword-plist-to-exclude-elements plist))
+            (expand-links (plist-get plist :expand-links))
             obj)
         (when only-element
           (narrow-to-region beg end))
@@ -1151,14 +1160,31 @@ etc.)."
           (setq obj (org-element-map obj org-element-all-elements
                       #'org-transclusion-content-filter-org-only-contents
                       nil nil '(section) nil)))
-        ;; For Org content, an extra +1 point seems to be added to the end of
-        ;; each element in the normalized content. This adds an extra empty
-        ;; line, which we would like to remove; hence content and src-end gets
-        ;; -1 here.
-        (list :src-content (substring (org-element-interpret-data obj) 0 -1)
+
+        ;; Expand file names in all the links
+        (when expand-links
+          (org-element-map obj 'link #'org-transclusion-content-filter-expand-links))
+
+        (list :src-content (org-element-interpret-data obj)
               :src-buf (current-buffer)
               :src-beg (point-min)
-              :src-end (1- (point-max)))))))
+              :src-end (point-max))))))
+
+(defun org-transclusion-content-filter-expand-links (link)
+  "Convert LINK to an absolute filename.
+LINK is assumed to be an Org element. This function does nothing
+to LINK if the link is already absolute.
+
+The current buffer is assumed to be the source buffer for the
+transclusion."
+  (when (string-equal "file" (org-element-property :type link))
+    (let ((raw-link (org-element-property :raw-link link)))
+      (unless (file-name-absolute-p raw-link)
+        (org-element-put-property
+         link :path
+         (expand-file-name
+          raw-link
+          (file-name-directory (buffer-file-name (current-buffer)))))))))
 
 (defun org-transclusion-content-filter-org-exclude-elements (data)
   "Exclude specific elements from DATA.
